// lorder.c
// list box ordering system
//    These function will implement a ordering system inside of a listbox.
// This will allow the user to select multiple lines and pick and insertion
// point to move the lines.
//    In order to use this system the follow steps must be followed.
//  
//       1. Include lorder.h inside the programs header file.  
//
//       2. Make sure the file insert.cur is in the program directory.  This
//          is the cursor that appears when the user selects an insertion 
//          point and it can be whatever the programmer wishes.
//
//       3. Add the the _LBOProc function to the EXPORTS list in the .DEF 
//          file.  The ordering system sub-classes the listbox and this is
//          the function
//
//       4. Add the lorder.c file to your program make file.
//
//       5. In the dialog callback functon call the function to implement 
//          the ordering.
//
//          To start ordering call the InitLBOSystem function with
//          the dialog handle and the control id.  This would normally
//          be in the WM_INITDIALOG message for each listbox needing
//          ordering.
//               i.e.     case WM_INITDIALOG:
//                           .
//                           InitLBOSystem(hInst,hDlg,LBC_NAMELIST);
//                           .
//                           break;
//
//       6. When creating the list box make sure that it does not have the 
//          LBS_SORT style and that it does have the LBS_MULTIPLESEL style.
//
//       7. Add the insert cursor to the rc file.  If you omit the line then
//          the system will use a north/south sizeing arrow.


// include files

#include "windows.h"
#include "lorder.h"


// structture definitions

typedef struct
   {
      short          first;                                          // flag for tracking first use

      short          hgt;                                            // hieght (in pixels) of each line
      short          div;                                            // division size for lines

      HCURSOR        cInsert;                                        // insert cursor
      HCURSOR        cNormal;                                        // normal cursor

      FARPROC        lpfunc;                                         // previous callback function
      short          CntrlID;                                        // control idea
   } _lbo_struct;


// function prototypes

LONG CALLBACK     _LBOProc       (HWND,UINT,WPARAM,LPARAM);          // callback function


// static variables

static   char     _szPropName[] = "LBOStruct";


// functions

void InitLBOSystem(HINSTANCE hInst,HWND hDlg,short id)
   {
      HWND                     hWnd;
      HANDLE                   hHandle;
      _lbo_struct far          *lpLBO;
      FARPROC                  lpfunc;
      
      if ((hWnd = GetDlgItem(hDlg,id)) != NULL)
         {
            if ((hHandle = GlobalAlloc(GMEM_ZEROINIT | GMEM_MOVEABLE,sizeof(_lbo_struct))) != NULL)
               {
                  lpLBO = (_lbo_struct far *)GlobalLock(hHandle);

                  lpLBO->first = TRUE;
                  if ((lpLBO->cInsert = LoadCursor(hInst,"insert")) == NULL)
                     {
                        lpLBO->cInsert = LoadCursor(NULL,IDC_SIZENS);
                     }
                  lpLBO->cNormal = LoadCursor(NULL,IDC_ARROW);

                  lpLBO->lpfunc = (FARPROC)GetWindowLong(hWnd,GWL_WNDPROC);
                  lpfunc = MakeProcInstance((FARPROC)_LBOProc,hInst);
                  SetWindowLong(hWnd,GWL_WNDPROC,(LONG)lpfunc);

                  lpLBO->CntrlID = id;
   
                  GlobalUnlock(hHandle);

                  SetProp(hWnd,_szPropName,hHandle);
               }
         }
   }


long CALLBACK _LBOProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
   {
      short                    my,result,ii,jj,count,destroy;
      short far                *iptr,*iptr2,*iptr3;   
      char                     q[80];
      long                     retval;
      DWORD                    ret;
      HANDLE                   hSel,hHandle;
      _lbo_struct far          *lpLBO;
      FARPROC                  lpfunc;

      retval = FALSE;

      if ((hHandle = GetProp(hWnd,_szPropName)) != NULL)
         {
            lpLBO = (_lbo_struct far *)GlobalLock(hHandle);

            destroy = FALSE;

            switch (message)
               {
                  case WM_LBUTTONDOWN:
                     if (lpLBO->first == FALSE)
                        {
                           my = HIWORD(lParam);
                           result = (my % lpLBO->hgt);
                           if (result <= lpLBO->div || result >= (lpLBO->hgt - lpLBO->div))
                              {
                                 if ((ret = SendMessage(hWnd,LB_GETSELCOUNT,0,0)) >0)
                                    {
                                       if ((hSel = GlobalAlloc(GMEM_ZEROINIT,sizeof(short) * ret)) != NULL)
                                          {
                                             iptr = (short far *)GlobalLock(hSel);

                                             SendMessage(hWnd,LB_GETSELITEMS,(WPARAM)ret,(LPARAM)(LPSTR)iptr);

                                             result = ((my + lpLBO->div) / lpLBO->hgt) + (short)SendMessage(hWnd,LB_GETTOPINDEX,0,0);

                                             count = (short)ret;
                                             for (ii=0, iptr2=iptr ; ii<count ; ii++, iptr2++)
                                                {
                                                   ret = *(iptr2) + ((*iptr2 < result) ? 0 : ii);
                                                   SendMessage(hWnd,LB_GETTEXT     ,(WPARAM)ret        ,(LPARAM)(LPSTR)q);
                                                   SendMessage(hWnd,LB_INSERTSTRING,(WPARAM)result + ii,(LPARAM)(LPSTR)q);
                                                }
                                             SendMessage(hWnd,LB_SETSEL      ,FALSE,MAKELPARAM(-1,0));
                                             SendMessage(hWnd,LB_SELITEMRANGE,TRUE ,MAKELPARAM(result,result + count - 1));

                                             for (ii=0, iptr2=iptr ; ii<count ; ii++, iptr2++)
                                                {
                                                   ret = *(iptr2) + ((*iptr2 < result) ? 0 : count);

                                                   for (jj=0, iptr3=iptr ; jj<ii ; jj++, iptr3++)
                                                      {
                                                         if ((*iptr3) <= (*iptr2))
                                                            {
                                                               ret--;
                                                            }
                                                      }
                                                   SendMessage(hWnd,LB_DELETESTRING,(WPARAM)ret,0);
                                                }

                                             GlobalUnlock(hSel);
                                             GlobalFree(hSel);
                                          }
                                    }
                                 else
                                    {
                                       retval = CallWindowProc(lpLBO->lpfunc,hWnd,message,wParam,lParam);
                                    }
                              }
                           else
                              {
                                 retval = CallWindowProc(lpLBO->lpfunc,hWnd,message,wParam,lParam);
                              }
                        }
                     break;

                  case WM_MOUSEMOVE:
                     if (lpLBO->first == TRUE)
                        {
                           lpLBO->hgt = (short)SendMessage(hWnd,LB_GETITEMHEIGHT,0,0);
                           lpLBO->div = (short)((double)lpLBO->hgt * .2);

                           lpLBO->first = FALSE;
                        }

                     if (SendMessage(hWnd,LB_GETSELCOUNT,0,0) > 0)
                        {
                           my = HIWORD(lParam);
                           result = my % lpLBO->hgt;

                           if (result <= lpLBO->div || result >= (lpLBO->hgt - lpLBO->div))
                              {
                                 SetCursor(lpLBO->cInsert);
                              }
                           else
                              {
                                 SetCursor(lpLBO->cNormal);
                              }
                        }
                     else
                        {
                           SetCursor(lpLBO->cNormal);
                        }

                     retval = CallWindowProc(lpLBO->lpfunc,hWnd,message,wParam,lParam);
                     break;

                  case WM_DESTROY:
                     lpfunc = (FARPROC)GetWindowLong(hWnd,GWL_WNDPROC);
                     SetWindowLong(hWnd,GWL_WNDPROC,(long)lpLBO->lpfunc);
                     FreeProcInstance(lpfunc);

                     if (lpLBO->cInsert != LoadCursor(NULL,IDC_SIZENS))
                        {
                           DestroyCursor(lpLBO->cInsert);
                        }

                     destroy = TRUE;

                     break;

                  default:
                     retval = CallWindowProc(lpLBO->lpfunc,hWnd,message,wParam,lParam);
                     break;
               }

            GlobalUnlock(hHandle);

            if (destroy == TRUE)
               {
                  GlobalFree(hHandle);
                  RemoveProp(hWnd,_szPropName);
               }
         }

      return(retval);
   }

